import { PublicKey } from "@solana/web3.js";
import { TransactionBuilder } from "@orca-so/common-sdk";
import { Context, PDA, RENEC_PROVIDER_TESTNET } from "..";
import { IsKycData, ProviderConfigData, UserConfigData, UserKycData } from "../types";

export class KycClient {
  ctx: Context;
  public pda: PDA;

  constructor(ctx: Context, pda: PDA) {
    this.ctx = ctx;
    this.pda = pda;
  }

  public static async getClient(ctx: Context): Promise<KycClient> {
    const pda = new PDA(ctx.program.programId);
    return new KycClient(ctx, pda);
  }

  public async createProviderConfig(
    provider: PublicKey,
    name: string,
  ): Promise<TransactionBuilder> {
    if (!name || name === "") {
      throw Error("Provider name is required");
    }
    if (name.length > 20) {
      throw Error("Invalid Provider name length, maximum 20 characters");
    }

    const providerConfig = this.pda.provider_config(name);

    const tx = (
      await this.ctx.methods.createProviderConfig({
        accounts: {
          authority: this.ctx.wallet.publicKey,
          providerConfigAccount: providerConfig.key,
        },
        inputs: {
          name,
          provider,
        },
      })
    ).toTx();

    return tx;
  }

  public async deactivateProvider(name: string): Promise<TransactionBuilder> {
    const providerConfig = this.pda.provider_config(name);
    const providerConfigData = await this.getProviderConfigByProviderName(name);
    if (providerConfigData.admin.toBase58() !== this.ctx.wallet.publicKey.toBase58()) {
      throw Error("Only Admin can deactivate provider");
    }

    const tx = (
      await this.ctx.methods.deactivateProvider({
        accounts: {
          authority: this.ctx.wallet.publicKey,
          providerConfigAccount: providerConfig.key,
        },
        inputs: {},
      })
    ).toTx();

    return tx;
  }

  public async deactivateUser(user: PublicKey): Promise<TransactionBuilder> {
    const userConfig = this.pda.user_config(user);
    const userConfigData = await this.getOneUserConfig(userConfig.key);
    if (userConfigData.admin.toBase58() !== this.ctx.wallet.publicKey.toBase58()) {
      throw Error("Only Admin can deactivate user");
    }

    const tx = (
      await this.ctx.methods.deactivateUser({
        accounts: {
          authority: this.ctx.wallet.publicKey,
          userConfig: userConfig.key,
        },
        inputs: {},
      })
    ).toTx();

    return tx;
  }

  public async getProviderConfigByProviderName(providerName: string): Promise<ProviderConfigData> {
    const pda = new PDA(this.ctx.program.programId);
    const providerConfig = pda.provider_config(providerName);

    const providerConfigData = await this.ctx.fetcher.getProviderConfig(providerConfig.key, true);
    if (!providerConfigData) {
      throw new Error(`Provider Config of provider name=${providerName} not found`);
    }
    return providerConfigData;
  }

  public async getUserConfig(user: PublicKey): Promise<UserConfigData> {
    const pda = new PDA(this.ctx.program.programId);
    const userConfig = pda.user_config(user);

    const userConfigData = await this.ctx.fetcher.getUserConfig(userConfig.key, true);
    if (!userConfigData) {
      throw new Error(`User Config of user ${user} not found`);
    }
    return userConfigData;
  }

  public async getOneUserConfig(userConfig: PublicKey): Promise<UserConfigData> {
    const userConfigData = await this.ctx.fetcher.getUserConfig(userConfig, true);
    if (!userConfigData) {
      throw new Error(`User Config ${userConfig} not found`);
    }
    return userConfigData;
  }

  public async getCurrentUserKyc(user: PublicKey): Promise<UserKycData> {
    const pda = new PDA(this.ctx.program.programId);
    const userConfig = pda.user_config(user);

    const userConfigData = await this.ctx.fetcher.getUserConfig(userConfig.key, true);
    if (!userConfigData) {
      throw new Error(`User Config of user ${user} not found`);
    }

    return this.getOneUserKyc(userConfigData.latestKycAccount);
  }

  public async getOneUserKyc(userKyc: PublicKey): Promise<UserKycData> {
    const userKycData = await this.ctx.fetcher.getOneUserKyc(userKyc, true);
    if (!userKycData) {
      throw new Error(`User Kyc ${userKyc} not found`);
    }

    const userConfigData = await this.getOneUserConfig(userKycData.userConfigAccount);
    if (userConfigData.provider.toBase58() === RENEC_PROVIDER_TESTNET) {
      userKycData.provider = "RENEC blockchain";
    } else {
      userKycData.provider = userConfigData.provider.toBase58();
    }

    return userKycData;
  }

  public async isKyc(user: PublicKey): Promise<IsKycData> {
    const pda = new PDA(this.ctx.program.programId);
    const userConfig = pda.user_config(user);

    const userConfigData = await this.ctx.fetcher.getUserConfig(userConfig.key, true);
    if (!userConfigData) {
      return {
        isKyc: false,
        kycLevel: null,
        isExpired: null,
      };
    }

    const userKycData = await this.getOneUserKyc(userConfigData.latestKycAccount);
    return {
      isKyc: true,
      kycLevel: userConfigData.kycLevel,
      isExpired: userKycData.isExpired,
    };
  }
}
