import * as anchor from "@project-serum/anchor";
import { ValidateKycData, Gender } from "../types";
import * as countryLookup from "country-code-lookup";
import { ProviderClient } from "../clients/provider-client";
const { createHash } = require("crypto");

export function keypairFromJson(secretKey: number[]): anchor.web3.Keypair {
  return anchor.web3.Keypair.fromSecretKey(Uint8Array.from(secretKey));
}

function hashNTimes(string: string | null, salt: string, n: number): string | null {
  if (!string || string === "") return null;
  let hash = createHash("sha256");
  let result = hash.update(string).update(salt).digest("hex");
  for (let i = 1; i < n; i++) {
    let hashInterval = createHash("sha256");
    hashInterval.update(result);
    result = hashInterval.digest("hex");
  }

  return result;
}

export const stringToBytes32 = (str: string | null): number[] => {
  if (!str) return [];
  const encoder = new TextEncoder();
  const encodedString = encoder.encode(str);
  const bytes32 = new Uint8Array(32);
  bytes32.set(encodedString.subarray(0, 32));
  const numbersArray: number[] = Array.from(bytes32, (byte) => byte);
  return numbersArray;
};

export const validateAndHashKycParams = (
  client: ProviderClient,
  salt: string,
  name: string,
  documentId: string,
  country: string,
  dob: Date,
  doe: Date | null,
  gender: Gender | null,
): ValidateKycData => {
  if (!salt || salt === "") {
    throw Error("Salt is required!");
  }
  if (salt.length > 20) {
    throw Error("Invalid salt length! Cannot exceed 20 characters");
  }
  if (name.length > 255 || documentId.length > 255) {
    throw Error("Invalid parameter length! Cannot exceed 255 characters");
  }
  if (country.length !== 2) {
    throw Error("Invalid country! Please provider valid country code");
  }

  let encodedCountry = country.toLowerCase();
  if (!countryLookup.byFips(encodedCountry)) {
    throw Error("Invalid country");
  }

  let encodedGender: number[] = [];
  if (gender) {
    if (!Object.values(Gender).includes(gender)) {
      throw Error("Invalid gender");
    }
    encodedGender = stringToBytes32(hashNTimes(gender, salt, client.genderHashRound));
  }

  return {
    encodedName: stringToBytes32(hashNTimes(name, salt, client.nameHashRound)),
    encodedDocumentId: stringToBytes32(hashNTimes(documentId, salt, client.documentIdHashRound)),
    encodedCountry,
    encodedDob: stringToBytes32(hashNTimes(dob.getTime().toString(), salt, client.dobHashRound)),
    encodedDoe: doe
      ? stringToBytes32(hashNTimes(doe.getTime().toString(), salt, client.doeHashRound))
      : [],
    encodedGender,
  };
};
